/* HPCU Bi-Weekly Challenge: MPI Pi - solution code 2 - uses MPI_Reduce()
 * Author: Aaron Weeden, Shodor, May 2015
 *
 * Approximate pi using a Left Riemann Sum under a quarter unit circle.
 *
 * When running the program, the number of rectangles can be passed using the
 * -r option, e.g. 'pi-mpi-2 -r X', where X is the number of rectangles.
 */

/*************
 * LIBRARIES *
 *************/
#include <stdio.h>  /* printf() */
#include <stdlib.h> /* exit(), EXIT_FAILURE */
#include <math.h>   /* sqrt() */
#include <float.h>  /* FLT_EPSILON and LDBL_DIG */
#include <mpi.h>    /* MPI_Init(), MPI_Send(), MPI_Recv(), etc. */
#include <unistd.h> /* getopt() */

/*************
 * CONSTANTS *
 *************/
#define TAG 0

/*************************
 * FUNCTION DECLARATIONS *
 *************************/
void setupMPI(int *argc, char ***argv, int *myRank, int *numProcs);

void getUserOptions(int argc, char **argv, const int myRank, int *numRects);

void distributeWork(const int numRects, const int myRank, const int numProcs,
    int *myNumRects, int *myDispl);

void syncData(const int myRank, const int numProcs, float *myArea);

/************************
 * FUNCTION DEFINITIONS *
 ************************/
int main(int argc, char **argv) {
  int numRects = 10;
  float width = 0.0;
  float x = 0.0;
  float heightSq = 0.0;
  float height = 0.0;
  float area = 0.0;
  float pi = 0.0;
  int i = 0;
  int myRank = 0;
  int numProcs = 1;
  int myNumRects = 0;
  int myDispl = 0;

  setupMPI(&argc, &argv, &myRank, &numProcs);

  getUserOptions(argc, argv, myRank, &numRects);

  distributeWork(numRects, myRank, numProcs, &myNumRects, &myDispl);

  width = (1.0 / numRects);

  for (i = myDispl; i < (myDispl + myNumRects); i++) {
    x = (i * width);

    heightSq = (1.0 - (x * x));

    /* Prevent nan value for sqrt() */
    if (heightSq < FLT_EPSILON) {
      heightSq = 0.0;
    }

    height = sqrt(heightSq);

    area += (width * height);
  }

  syncData(myRank, numProcs, &area);

  if (myRank == 0) {
    pi = (4.0 * area);

    printf("%.*f\n", LDBL_DIG, pi);
  }

  MPI_Finalize();

  return 0;
}

void setupMPI(int *argc, char ***argv, int *myRank, int *numProcs) {
  MPI_Init(&(*argc), &(*argv));
  MPI_Comm_rank(MPI_COMM_WORLD, &(*myRank));
  MPI_Comm_size(MPI_COMM_WORLD, &(*numProcs));
}

void getUserOptions(int argc, char **argv, const int myRank, int *numRects) {
  char c;

  while ((c = getopt(argc, argv, "r:")) != -1) {
    switch(c) {
      case 'r':
        (*numRects) = atoi(optarg);
        break;
      case '?':
      default:
        if (myRank == 0) {
          fprintf(stderr, "Usage: ");
          fprintf(stderr, "%s [-r numRects]\n", argv[0]);
        }
        MPI_Finalize();
        exit(EXIT_FAILURE);
    }
  }
  argc -= optind;
  argv += optind;
}

/* Split rectangles as evenly as possible, with the first N processes getting 1
 * extra rectangle */
void distributeWork(const int numRects, const int myRank, const int numProcs,
    int *myNumRects, int *myDispl) {
  int evenSplit = (numRects / numProcs);
  int numProcsWith1Extra = (numRects % numProcs) + 1;

  if (numProcsWith1Extra == 1) {
    numProcsWith1Extra = 0;
  }

  if (myRank < numProcsWith1Extra) {
    (*myNumRects) = (evenSplit + 1);
    (*myDispl) = (myRank * (evenSplit + 1));
  }
  else {
    (*myNumRects) = evenSplit;
    (*myDispl) = (numProcsWith1Extra * (evenSplit + 1)) +
      ((myRank - numProcsWith1Extra) * evenSplit);
  }
}

void syncData(const int myRank, const int numProcs, float *area) {
  float tmp = (*area);
  MPI_Reduce(&tmp, &(*area), 1, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
}
